/*
 * utilBag.c: Functions for managing bags. Extends hash table functionality.
 *
 * 
 */
#include "utilBag.h"

/* PRIMITIVE: CREATE: No separate function. Just call createHashTable()
with an initKeyFunc that expects to be passed a bagRec; shouldFree must be
true */


/* PRIMITIVE: ADD: adds a new object with a count of 1. If it is already there, 
adds 1 to the count. returns NULL on failure. Normally returns the number of items (>=1) */
int addToBag(theTable, item, numTimes)
hashTable theTable;
itemPtr item;
int numTimes;
  {
    bagRec addResult;
    bagRec newRecord;
    char *(*realMallocFunc)();
    void (*realFreeFunc)();

    realMallocFunc = (theTable->mallocFunc == NULL) ? DEFAULT_MALLOC : theTable->mallocFunc;
  realFreeFunc = (theTable->freeFunc == NULL) ? DEFAULT_FREE : theTable->freeFunc;

    /* create and initialize the new bagrec */
    newRecord = (bagRec)(realMallocFunc(sizeof(struct bagRecStruct)));
    if(newRecord == NULL) return(NULL);

    newRecord->data = item;
    newRecord->count = numTimes;

    /* add the bagrec or update the count of an existing one */ 
    addResult = (bagRec)addToHashTable(theTable, newRecord);
    if(addResult == NULL)
    {
      realFreeFunc(newRecord);
      return(NULL);
    }

    if(addResult != newRecord)
    {
      /* the item already existed */
     realFreeFunc(newRecord);
     addResult->count += numTimes;
     return addResult->count;
   }
 return numTimes;
}

/* PRIMITIVE: FIND: No separate function. Just use findInHashTable and expect a
bagRec in return */

/* PRIMITIVE: DELETE: deletes a number of occurrences of an item (until there
are zero left). A negative number returned indicates that the request was to
delete more items than existed. Note: This is not maximally efficient. */

int deleteFromBag(theTable, item, numTimes)
hashTable theTable;
itemPtr item;
int numTimes;
  {
    bagRec foundRec;
    int returnCount;
    char *(*realKeyFunc)();
    void (*realFreeFunc)();

    realKeyFunc = (theTable->keyFunc == NULL) ? DEFAULT_KEYFUNC : theTable->keyFunc;
    realFreeFunc = (theTable->freeFunc == NULL) ? DEFAULT_FREE : theTable->freeFunc;

    foundRec = (bagRec)findInHashTable(theTable, realKeyFunc(item));
    if(foundRec == NULL)
      {
        /* no such item ever existed */
        return(- numTimes);
      }
    
    /* we now have found a record */
    if(numTimes >= foundRec->count)
      {
        /* the item should be completely deleted */
        returnCount = foundRec->count - numTimes;
        deleteFromHashTable(theTable, foundRec);
        realFreeFunc(foundRec);
        return(returnCount);
      }
    
    /* we now know all we have to do is decrement the count */
    foundRec->count -= numTimes;

    return(foundRec->count);
}


/* PRIMITIVE DESTROY: Use destroyHashTable */

/* PRIMITIVE: MAP: Use hashTableDo with a function that expects a bagRec */

/* PRIMITIVE: SORTMAPANDDESTROY: Goes through the hash table in sorted
order by count and then destroys it. The destruction is necessary because
the hash table is physically rearranged. This is useful for processing a
histogram. The function should expect the most frequest items first */

void sortMapAndDestroyBag(theTable, aFunc, maxCases, extraArg)
hashTable theTable;
void (*aFunc)(); /* takes an item (pointer) as first arg and an arbitrary second arg */
char *extraArg; /* arbitrary second arg to be passed to aFunc */
int maxCases; /* the maximum of elements in the histogram to return */
{
  int i, limit;
  itemPtr *currentItem, elementsArray;
  if(theTable->numHashed > 0)
    {
      /* sort the table, thus destroying its hashed structure */
      /* we actually the array of pointers (or pseudo-pointers such as NULL)
      elementsArray = theTable->elements;
      qsort(elementsArray, theTable->size, sizeof(char *), countSort); 

      /* enumerate the table stopping when we reach only deleted elements */
      currentItem = theTable->elements;
      limit = theTable->size > maxCases ? maxCases : theTable->size;
      for(i=0;
          i<limit
          && *currentItem != NULL
          && *currentItem !=HASH_DELETED;
        i++)
        {
          (aFunc)(*currentItem,extraArg);
          currentItem++;
        }
    }
  /* destroy the table */
  destroyHashTable(theTable);
}

/* function to sort a bash table of bagrecs ensuring that the largest count
is first and all NULL or deleted elements are at the end */

int countSort(el1, el2)
bagRec el1;
bagRec el2;
{
  if(el1 == NULL || (char *)el1 == HASH_DELETED)
    {
      if(el2 == NULL || (char *)el2 == HASH_DELETED)
        {
          return(0);
        }
      else
        {
          return(1); /* first element null goes at the end */
        }
    }
  else
    {
      if(el2 == NULL || (char *)el2 == HASH_DELETED)
        {
          return(-1); /* second element null means first element goes before */
        }
      else
        {
          /* both elements are real pointers */
          return(el2->count - el1->count);
        }
    }
}


/* UTILITY for when the items we are entering are just character strings with no
other data. This is an example of what could be passed to createHashTable as the second argument. The items actually compared are bagrecs */

char *testGetBagRec(item)
bagRec item;
{
  return((char *)(item->data));
}
